// Primary widgets
@import 'widgets/alttab';
@import 'widgets/applets';
@import 'widgets/base';
@import 'widgets/buttons';
@import 'widgets/calendar';
@import 'widgets/desklet';
@import 'widgets/dialogs';
@import 'widgets/screensaver';
@import 'widgets/expo';
@import 'widgets/ibus-popup';
@import 'widgets/keyboard';
@import 'widgets/menus';
@import 'widgets/notifications';
@import 'widgets/osd';
@import 'widgets/panel';
@import 'widgets/startmenu';
@import 'widgets/switch-check';
@import 'widgets/windowlist';
@import 'widgets/deprecated';
