// desklet

.desklet {
  &:highlight {
    border: 2px solid $accent_color;
    border-radius: $base_border_radius;
  }

  &-with-borders {
    border: 1px solid $borders_color;
    border-radius: $base_border_radius;
    background-color: transparentize($bg_color, 0.2);
  }

  &-with-borders-and-header {
    border: 1px solid $borders_color;
    border-radius: 0 0 $base_border_radius $base_border_radius;
    border-top-width: 0;
    background-color: transparentize($bg_color, 0.2);
    padding: $base_padding * 2;
  }

  &-header {
    border: 1px solid $borders_color;
    border-radius: $base_border_radius $base_border_radius 0 0;
    background-color: transparentize($bg_color, 0.2);
    padding: $base_padding;
  }

  &-drag-placeholder {
    border: 1px solid red;
  }
}

// digital photo frame

.photoframe-box {
  border: 1px solid $borders_color;
  border-radius: $base_border_radius;
  background-color: transparentize($bg_color, 0.2);
  padding: $base_padding;
}

// clock

.clock-desklet-label {}
