// global colors
$bg_color: #242428;
$base_color: #303034;
$fg_color: white;

$panel_bg: #1a1a1f;

// colored elements
$error_color: #e4705b;
$error_bg_color: #c01c28;

$success_color: #8ff0a4;
$success_bg_color: #26a269;

$destructive_bg_color: #c01c28;
$destructive_color: #ff7b63;

$warning_color: #f8e45c;
$warning_bg_color: #cd9309;

$notification_badge_bg_color: #e74b37;

$accent_color: #78aeed;
$accent_bg_color: #3584e4;

$large_icon_color: transparentize($fg_color, 0.6);
$light_text_color: mix($fg_color, $bg_color, 65%);

// used for button, entries and highlighted items
$light_bg_color: #39393c;
$light_base_color: #444448;
$lighter_bg_color: #454549;
$lightest_bg_color: #656569;

// link colors
$link_color: #1c71d8;

// borders
$borders_color: #333337;

// shadows
$shadow_color: rgba(0, 0, 0, 0.5);

// insensitive state
$insensitive_fg_color: mix($fg_color, $bg_color, 40%);
