/* -*- mode: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*- */
#ifndef __CINNAMON_APP_SYSTEM_PRIVATE_H__
#define __CINNAMON_APP_SYSTEM_PRIVATE_H__

#include "cinnamon-app-system.h"

void _cinnamon_app_system_notify_app_state_changed (CinnamonAppSystem *self, CinnamonApp *app);

#endif
